function [particleAvg particleEval allParticles] = particleFilter(configFile,LDA,psi,recallList,allParticles)
% This function implements the particle filter, see paper for details.
% Input:
%   configFile
%   LDA - fitted topic-word matrix
%   psi - study list
%   recallList - list of recalled words
%   allParticles - datastructure that holds all the initialized particles
% Output:
%   particleAvg - all parameters as averaged from all particles
%   particleEval - APE and ESS of all particles across subjects
%   allParticles - particles after learning
% Author: Richard Socher

eval(configFile);

[K W] = size(LDA.beta); %word distribution matrix
sbeta = softmax(LDA.beta');

psi_lda_numbers = getLDAWordNumbers(psi,LDA.vocab);

studyListLength = length(psi);
recallListLength = length(recallList);
numOfParticles = length(allParticles);
particleEval.APE = zeros(1,recallListLength);
particleEval.APE_intrusion = zeros(1,recallListLength);
particleEval.APE_chance_onlyStudyWords = zeros(1,recallListLength);
particleEval.APE_chance_allLDAwords = zeros(1,recallListLength);

particleEval.ESS = zeros(1,recallListLength);

%uniform topic prior probabilities in mean parameter space
pr = ones(K,1)./K;
% stheta is in mean parameter space, normal theta is in log space
for c = 1:length(allParticles)
    allParticles{c}.stheta = softmax(allParticles{c}.theta);
    %initial context is last study context
    allParticles{c}.stheta(:,end+1) = allParticles{c}.stheta(:,end);
    allParticles{c}.theta(:,end+1) = allParticles{c}.theta(:,end);
end

totalTimeSeriesLength =(studyListLength+1)+recallListLength;
for t = (studyListLength+2):totalTimeSeriesLength
    currentWordIndex = t-(studyListLength+1);
    
    for c = 1:numOfParticles
        %parameters of this particle
        lambda = allParticles{c}.subjectParameters.lambda;
        eps1 = allParticles{c}.subjectParameters.eps1;
        
        %----------------------
        % 1. get posterior of words given contexts to compute weight for each particles
        %----------------------
        delta_s = zeros(W,1);
        
        % if we recall episodically (E, ES), if S, then this is ignored
        if (1-lambda)> 0% speed up if:
            for j = 1:length(psi_lda_numbers)
                dist = distance(allParticles{c}.stheta(:,t), allParticles{c}.stheta(:,j),InferParam.distanceMetric);
                newWeight = 1./(dist^eps1);
                numOfTimesThisWordHasBeenRecalled = sum(recallList(1:(currentWordIndex-1)) == psi_lda_numbers(j));
                % the jth word of the study list has been recalled and is hence weighted down
                if numOfTimesThisWordHasBeenRecalled>0
                    % this word has already been recalled and is hence weighted down
                    newWeight = (InferParam.lowerRecalledWordWeight)^numOfTimesThisWordHasBeenRecalled * newWeight;
                end
                delta_s(psi_lda_numbers(j)) = newWeight;
            end
        end
        
        switch lambda
            case 0
                % E: purely episodic recall
                if sum(delta_s)>0
                    delta_s = delta_s./sum(delta_s);
                    delta_s = [delta_s ;zeros(length(LDA.vocab) - length(delta_s),1)];
                    phi = delta_s;
                    if ~(abs(sum(phi)-1)<0.00000001)
                        error(['multinomial parameters do not sum to 1, sum(phi)=' num2str(sum(phi))])
                    end
                    
                    allParticles{c}.weight = phi(recallList(currentWordIndex))+eps;
                    allParticles{c}.phi = phi+eps;
                else
                    % special case, for intrusions after all study words
                    disp('SPECIAL: All study words have been recalled, now an intrusion (E)!')
                    allParticles{c}.weight = eps;
                    allParticles{c}.phi = eps;
                end
            case 1
                %S: purely Semantic recall, e.g. delta is zero and we may recall
                % semantically or lambda == 1 and deltas are ignored.
                oldTheta = allParticles{c}.stheta(:,t);
                semanticRecall =sbeta*oldTheta;
                
                semanticRecall = [semanticRecall ;zeros(length(LDA.vocab) - length(semanticRecall),1)];
                phi = semanticRecall;
                if ~(abs(sum(phi)-1)<0.00000001)
                    error(['multinomial parameters do not sum to 1, sum(phi)=' num2str(sum(phi))])
                end
                allParticles{c}.weight = phi(recallList(currentWordIndex))+eps;
                allParticles{c}.phi = phi+eps;
            otherwise
                % SE: Semantic and episodic recall
                if sum(delta_s)>0
                    delta_s = delta_s./sum(delta_s);
                    delta_s = [delta_s ;zeros(length(LDA.vocab) - length(delta_s),1)];
                else
                    % all words have been recalled, switching to purely
                    % semantic recall and ignore delta
                    disp('All study words have been recalled, now an intrusion (SE)!')
                    delta_s = 0;
                    lambda = 1;
                end
                
                oldTheta = allParticles{c}.stheta(:,t);
                semanticRecall = sbeta*oldTheta; % this takes most of the time in inference, since sbeta is so large :/
                
                semanticRecall = [semanticRecall ;zeros(length(LDA.vocab) - length(semanticRecall),1)];
                
                phi = lambda.*semanticRecall + (1-lambda).*delta_s;
                if ~(abs(sum(phi)-1)<0.000001)
                    error(['multinomial parameters do not sum to 1, sum(phi)=' num2str(sum(phi))])
                end
                allParticles{c}.weight = phi(recallList(currentWordIndex))+eps;
                allParticles{c}.phi = phi+eps;
        end
        
    end
    
    
    %----------------
    %APE, ESS, APE_chance
    phiSum= zeros(length(phi),1);
    for c = 1:numOfParticles
        phiSum = phiSum + allParticles{c}.phi;
    end
    phiAvg = phiSum/numOfParticles;
    phiAvg = 1./sum(phiAvg) * phiAvg;
    
    probFromModel = phiAvg(recallList(currentWordIndex));
    if probFromModel==0
        probFromModel = eps;
    end
    PE = -log(probFromModel);
    particleEval.APE(currentWordIndex) = PE;
    
    if(~any(psi_lda_numbers == recallList(currentWordIndex)))
        disp(['Saving intrusion APE: ' num2str(particleEval.APE(currentWordIndex))])
        particleEval.APE_intrusion(currentWordIndex) = particleEval.APE(currentWordIndex);
        particleEval.APE_chance_onlyStudyWords(currentWordIndex) = -log(eps);
    end
    
    % chance models
    particleEval.APE_chance_allLDAwords(currentWordIndex) = -log(1./W);
    if isfield(particleEval,'APE_intrusion')
        particleEval.APE_chance_onlyStudyWords(currentWordIndex) = -log(1./(studyListLength-currentWordIndex+1-sum(particleEval.APE_intrusion~=0)));
    else
        particleEval.APE_chance_onlyStudyWords(currentWordIndex) = -log(1./(studyListLength-currentWordIndex+1));
    end
    
    %----------------------
    % 2. normalize weights
    %----------------------
    particleWeightSum = 0;
    for c = 1:numOfParticles
        particleWeightSum = particleWeightSum + allParticles{c}.weight;
    end
    
    particleResampleMult = zeros(1,numOfParticles);
    for c = 1:numOfParticles
        if particleWeightSum==0
            % in case all particles gave 0 probability to a word
            allParticles{c}.weight = 1./numOfParticles;
        else
            allParticles{c}.weight = (allParticles{c}.weight / particleWeightSum);
        end
        particleResampleMult(c) = allParticles{c}.weight;
    end
    
    % compute effective sample size according to doucet
    % see: A tutorial on pf & smooth, 15 years later, p14
    essNow = 1./(sum(particleResampleMult.*particleResampleMult));
    particleEval.ESS(currentWordIndex) = essNow;
    
    if(~any(psi_lda_numbers == recallList(currentWordIndex)))
        particleEval.ESS_intrusion(currentWordIndex) = particleEval.ESS(currentWordIndex);
    end
    
    %----------------------
    % 3. resample particles
    %----------------------
    sampledThetas = mnrnd(numOfParticles,particleResampleMult,1);
    
    i = 1;
    newParticles = cell(length(allParticles),1);
    for c = 1:numOfParticles
        for newC = 1:sampledThetas(c)
            newParticles{i} = allParticles{c};
            i=i+1;
        end
    end
    allParticles = newParticles;
    
    
    %----------------------
    % 4. drift new particles for next time step
    %----------------------
    for c = 1:numOfParticles
        
        eta2 = allParticles{c}.subjectParameters.eta_r(1);
        eta3 = allParticles{c}.subjectParameters.eta_r(2);
        Sigma_r = allParticles{c}.subjectParameters.sigma*eye(K);
        
        if recallList(currentWordIndex)>LDA.oriVocabEnd
            % this word did not appear in the original LDA matrix and hence
            % gets only a uniform topic distribution
            p_r = 1./K .* ones(K,1);
        else
            p_r = exp(LDA.beta(:,recallList(currentWordIndex)));
        end
        
        p_r = p_r +  eps;                     %avoid -\infty if prob of a word is 0...
        p_r = (p_r ./ sum(p_r)) .* pr;        %normalize
        
        %sum of weighted study thetas
        match_ind = psi_lda_numbers == recallList(currentWordIndex);
        if(~any(match_ind))
            % word intrusion! (renormalize, so both components sum to one)
            h = (eta2/(eta2+eta3))*allParticles{c}.theta(:,t) + (eta3/(eta2+eta3))*log(p_r);
        else
            %new mean of drifting context, see Eq. (2) in the report
            recall_theta_study = allParticles{c}.theta(:,match_ind);
            h = eta2*allParticles{c}.theta(:,t) + eta3*log(p_r) + (1-eta2-eta3)*recall_theta_study;
        end
        %sample new context
        allParticles{c}.theta(:,t+1) = mvnrnd(h,Sigma_r);
        allParticles{c}.stheta(:,t+1) = softmax(allParticles{c}.theta(:,t+1));
        
        if sum(exp(allParticles{c}.theta(:,t+1))) == 0
            error('exp(theta) = 0.')
        end
        
        if ~(sum(allParticles{c}.theta(:,t+1)) > -Inf)
            error('Theta = -Inf.')
        end
    end
    
end


%----------------------
% 5. average particles' thetas and parameters
%----------------------
sthetaSum = zeros(K,totalTimeSeriesLength+1);

% get sum of parameters
sigmaSum =0;
lambdaSum =0;
eta_s1Sum =0;
eps1Sum =0;
eta_rSum = zeros(1,3);

for c = 1:numOfParticles
    sthetaSum = sthetaSum+ allParticles{c}.stheta;
    
    eta_s1Sum = eta_s1Sum + allParticles{c}.subjectParameters.eta_s1;
    eta_rSum = eta_rSum + allParticles{c}.subjectParameters.eta_r;
    sigmaSum = sigmaSum + allParticles{c}.subjectParameters.sigma;
    lambdaSum = lambdaSum + allParticles{c}.subjectParameters.lambda;
    eps1Sum = eps1Sum + allParticles{c}.subjectParameters.eps1;
end

% get expected value
particleAvg.allThetas_sr = sthetaSum/numOfParticles;

particleAvg.subjParam.eta_s1 = eta_s1Sum/numOfParticles;
particleAvg.subjParam.eta_r = eta_rSum/numOfParticles;
particleAvg.subjParam.sigma = sigmaSum/numOfParticles;
particleAvg.subjParam.lambda =lambdaSum/numOfParticles;
particleAvg.subjParam.eps1 =eps1Sum/numOfParticles;

% get variance
sigmaSumVar =0;
lambdaSumVar =0;
eta_s1SumVar =0;
eps1SumVar =0;
eta_rSumVar = zeros(1,3);
for c = 1:numOfParticles
    eta_s1SumVar = eta_s1SumVar + (allParticles{c}.subjectParameters.eta_s1 - particleAvg.subjParam.eta_s1 )^2;
    eta_rSumVar = eta_rSumVar + (allParticles{c}.subjectParameters.eta_r - particleAvg.subjParam.eta_r).^2;
    sigmaSumVar = sigmaSumVar + (allParticles{c}.subjectParameters.sigma - particleAvg.subjParam.sigma)^2;
    lambdaSumVar = lambdaSumVar + (allParticles{c}.subjectParameters.lambda - particleAvg.subjParam.lambda)^2;
    eps1SumVar = eps1SumVar + (allParticles{c}.subjectParameters.eps1 - particleAvg.subjParam.eps1)^2;
end

particleAvg.subjParam.var.eta_s1 = eta_s1SumVar/(numOfParticles-1); % -1 to be unbiased ^^
particleAvg.subjParam.var.eta_r = eta_rSumVar/(numOfParticles-1);
particleAvg.subjParam.var.sigma = sigmaSumVar/(numOfParticles-1);
particleAvg.subjParam.var.lambda =lambdaSumVar/(numOfParticles-1);
particleAvg.subjParam.var.eps1 =eps1SumVar/(numOfParticles-1);
