function [psi theta_s] = sim_study(prm,LDA,list)

%Simulate study phase. An initial context is drawn from a zero-mean
%multivariate Gaussian with isotropic unit variance.
%
%USAGE: [psi theta_s] = sim_study(beta,pr,list,eta_s,sigma_s)
%
%INPUTS:
%   model - gtcm model structure
%   pr - [K x 1] topic priors
%   list - [1 x N] indices (corresponding to columns of beta) of study words
%   eta_s - gain parameter on posterior
%   sigma_s - drift variance
%
%OUTPUTS:
%   psi - [1 x W] empirical distribution (delta functions at study words)
%   theta_s - [K x N] matrix of study contexts
%
%EXAMPLE:
%   beta = load('data/tasa/final.beta');  %load beta matrix from LDA on the NYT database
%   pr = 1./ones(size(beta,1),1);                       %uniform prior over topics
%   list = 1:10;                                       %construct a word list
%   lda_nounpool = lower(textread('data/tasa/vocab.dat','%s'));
%   [psi theta_s] = sim_study(beta,pr,list,0.5,1,lda_nounpool);      %run simulation
%
%Sam Gershman, Richard Socher
%April 25, 2009

K = size(LDA.beta,1); %word distribution matrix (number of topics)
N = length(list);     %number of study words

%parameters
eta1 = prm.eta1;
sigma = prm.sigma;
Sigma_s = sigma*eye(K);

%initial context
theta_s = zeros(K, N+1);
theta_s(:,1) = mvnrnd(zeros(1,K),eye(K));

%uniform topic prior probabilities in mean parameter space
pr = ones(K,1)./K;

%loop through study words
for n = 1:N
    if list(n)>LDA.oriVocabEnd
        % this word did not appear in the original LDA matrix and hence
        % gets only a uniform topic distribution
        p = 1./K .* ones(K,1);
    else
        p = exp(LDA.beta(:,list(n)));               %beta is in log space, so either stay in log space and add log(pr), or take exp
    end
    p = p +  eps;                               %avoid -\infty if prob of a word is 0...
    p = (p ./ sum(p)) .* pr;                            %posterior over topics
    m = eta1*theta_s(:,n) + (1-eta1)*log(p);    %mean of drift
    theta_s(:,n+1) = mvnrnd(m,Sigma_s);         %sample new context
end

%empirical distribution
psi = LDA.vocab(list);
